unit data;

{$mode objfpc}{$H+}

interface

uses
	Classes, SysUtils, fgl;

type

	TFPGIntegers = specialize TFPGList<Integer>;

	{ TDataTable }

	generic TDataTable<T> = class
	public type

		TDataSlot = record
			Value: T;
			Used: Boolean;
		end;

		{ TDataTableEnumerator }

		TDataTableEnumerator = object
		public
			function GetCurrent: T;
			function MoveNext: Boolean;
			procedure Reset;
		public
			property Current: T read GetCurrent;
		private
			FTable: TDataTable;
			FCurrentIndex: Integer;
		public
			constructor Init(ATable: TDataTable);
		end;
	private
		FSlots: array of TDataSlot;
		FUsedSlots: TFPGIntegers;
	private
		procedure SetValue(AIndex: Integer; AValue: T);
		function GetValue(AIndex: Integer): T;
		function GetSize: Integer;
		procedure SetSize(ASize: Integer);
		function GetUsedValue(AIndex: Integer): T;
		function GetUsedValuesCount: Integer;
	public
		property UsedValues[AIndex: Integer]: T read GetUsedValue;
		property UsedValuesCount: Integer read GetUsedValuesCount;
		property Values[AIndex: Integer]: T read GetValue write SetValue; default;
		property Size: Integer read GetSize write SetSize;
	public
		constructor Create;
		destructor Destroy; override;
		procedure Clear;
		function GetValue(AIndex: Integer; out AValue: T): Boolean; inline;
		function GetEnumerator: TDataTableEnumerator;
	end;

	{ TDataTableManager }

	generic TDataTableManager<T> = class
	public type
		TTable = specialize TDataTable<T>;
	private
		FTableSize: Integer;
	public
		function CreateInstance: TTable;
		procedure ClearInstance(AInstance: TTable);
		procedure DestroyInstance(AInstance: TTable);
	public
		constructor Create(ATableSize: Integer);
		destructor Destroy; override;
	end;

implementation

{ TDataTable.TDataTableEnumerator }

function TDataTable.TDataTableEnumerator.GetCurrent: T; inline;
begin
	Result:= FTable.UsedValues[FCurrentIndex];
end;

function TDataTable.TDataTableEnumerator.MoveNext: Boolean; inline;
begin
	Inc(FCurrentIndex);
	Result:= FCurrentIndex < FTable.UsedValuesCount;
end;

procedure TDataTable.TDataTableEnumerator.Reset; inline;
begin
	FCurrentIndex:= -1;
end;

constructor TDataTable.TDataTableEnumerator.Init(ATable: TDataTable);
begin
	FTable:= ATable;
	Reset;
end;

{ TDataTableManager }

function TDataTableManager.CreateInstance: TTable;
begin
	Result:= TTable.Create;
	Result.Size:= FTableSize;
end;

procedure TDataTableManager.ClearInstance(AInstance: TTable);
begin
	AInstance.Clear;
end;

procedure TDataTableManager.DestroyInstance(AInstance: TTable);
begin
	AInstance.Destroy;
end;

constructor TDataTableManager.Create(ATableSize: Integer);
begin
	FTableSize:= ATableSize;
end;

destructor TDataTableManager.Destroy;
begin
	inherited Destroy;
end;

{ TDataTable }

procedure TDataTable.SetValue(AIndex: Integer; AValue: T); inline;
begin
	with FSlots[AIndex] do begin
		Value:= AValue;

		if not Used then begin
			Used:= True;
			FUsedSlots.Add(AIndex);
		end;
	end;
end;

function TDataTable.GetValue(AIndex: Integer; out AValue: T): Boolean; inline;
begin
	with FSlots[AIndex] do begin
		AValue:= Value;
		Result:= Used;
	end;
end;

function TDataTable.GetValue(AIndex: Integer): T; inline;
begin
	if not GetValue(AIndex, Result) then begin
		{$ifopt D+}
		raise Exception.Create('Uninitialized slot accessed.');
		{$endif}
	end;
end;

function TDataTable.GetSize: Integer; inline;
begin
	Result:= Length(FSlots);
end;

procedure TDataTable.SetSize(ASize: Integer); inline;
begin
	SetLength(FSlots, ASize);
end;

function TDataTable.GetUsedValue(AIndex: Integer): T; inline;
begin
	Result:= FSlots[FUsedSlots[AIndex]].Value;
end;

function TDataTable.GetUsedValuesCount: Integer; inline;
begin
	Result:= FUsedSlots.Count;
end;

constructor TDataTable.Create;
begin
	FSlots:= nil;
	FUsedSlots:= TFPGIntegers.Create;
end;

destructor TDataTable.Destroy;
begin
	FUsedSlots.Destroy;
	inherited;
end;

procedure TDataTable.Clear; inline;
var
	i: Integer;
begin
	for i in FUsedSlots do begin
		FSlots[i].Used:= False;
	end;

	FUsedSlots.Count:= 0;
end;

function TDataTable.GetEnumerator: TDataTableEnumerator; inline;
begin
	Result.Init(Self);
end;

end.

